#!/usr/bin/env python

import nsysstats

class AsyncMemcpyPageable(nsysstats.Report):

    ROW_LIMIT = 50

    usage = f"""{{SCRIPT}}[:<rowlimit>] -- AsyncMemcpyPageable

    <rowlimit> - Maximum number of rows returned by the query. If not given,
        only the top {ROW_LIMIT} results are displayed.

    Output:
        API Duration : Duration of runtime API in nanoseconds
        API Start Time : Start time of runtime API in nanoseconds
        Duration on GPU : Duration of memcpy on GPU in nanoseconds
        Memcpy Start Time : Start time of memcpy on GPU in nanoseconds
        Source Kind : Memcpy source memory kind
        Destination kind : Memcpy destination memory kind
        Bytes : Number of bytes transferred
        Stream ID : Stream identifier
        API Name : Name of runtime API function

    This rule identifies asynchronous memory transfers that
    end up becoming synchronous if the memory is pageable.
"""

    query_stub = """
WITH
    {MEM_KIND_STRS_CTE}
    runtime AS (
        SELECT
            start,
            end - start AS duration,
            correlationId,
            nameId
        FROM
            CUPTI_ACTIVITY_KIND_RUNTIME
    ),
    async AS (
        SELECT
            id,
            value
        FROM
            StringIds
        WHERE
            value LIKE '%cpy%async%'
    ),
    memcpy AS (
        SELECT
            srcKind,
            dstKind,
            start,
            end - start AS duration,
            bytes,
            streamId,
            correlationId
        FROM
            CUPTI_ACTIVITY_KIND_MEMCPY
        WHERE
            srcKind = 0 OR dstKind = 0
    )
    SELECT
        runtime.duration AS "API Duration (ns)",
        runtime.start AS "API Start Time (ns)",
        memcpy.duration AS "Duration on GPU (ns)",
        memcpy.start AS "Memcpy Start Time (ns)",
        msrck.name AS "Source Kind",
        mdstk.name AS "Destination Kind",
        memcpy.bytes AS "Bytes",
        memcpy.streamId AS "Stream ID",
        async.value AS "API Name"
    FROM ((runtime
    INNER JOIN async
        ON async.id = runtime.nameId)
    INNER JOIN memcpy
        ON runtime.correlationId = memcpy.correlationId)
    LEFT JOIN
        MemKindStrs AS msrck
        ON memcpy.srcKind = msrck.id
    LEFT JOIN
        MemKindStrs AS mdstk
        ON memcpy.dstKind = mdstk.id
    ORDER BY cast(runtime.duration as int) DESC
    LIMIT {ROW_LIMIT}
;
"""

    table_checks = {
        'CUPTI_ACTIVITY_KIND_RUNTIME':
            'There were no problems detected because the report does not'
            ' contain CUDA trace data.',
        'CUPTI_ACTIVITY_KIND_MEMCPY':
            'There were no problems detected because the report does not'
            ' contain CUDA trace data.',
    }

    def setup(self):
        err = super().setup()
        if err != None:
            return err

        if self.args and self.args[0].isdigit():
            row_limit = self.args[0]
        else:
            row_limit = self.ROW_LIMIT

        self.query = self.query_stub.format(
            MEM_KIND_STRS_CTE = self.MEM_KIND_STRS_CTE,
            ROW_LIMIT = row_limit)

if __name__ == "__main__":
    AsyncMemcpyPageable.Main()
