#!/usr/bin/env python

import nsysstats

class CUDAGPUStarvation(nsysstats.Report):

    THRESHOLD = 500
    ROW_LIMIT = 50

    usage = f"""{{SCRIPT}}[:<rowlimit>[:<threshold>]] -- GPU Starvation (kernels + memory operations)

    <rowlimit> - Maximum number of rows returned by the query. If not given,
        only the top {ROW_LIMIT} results are displayed.
    <threshold> - Minimum duration of GPU gaps in milliseconds. If not given,
        only gaps longer than {THRESHOLD}ms are returned.

    Output:
        Row# : Row number of GPU gap
        Duration : Duration of GPU gap in nanoseconds
        Start time : Start time of GPU gap in nanoseconds
        PID : Process identifier
        Device : GPU device name

    This rule identifies ranges where a GPU is idle for more than a set threshold.
    GPU gaps that cannot be addressed by the user are excluded. This includes:
        1. CUDA profiling overhead in the middle of a GPU gap.
        2. The initial gap starting before the first GPU operation, which
           corresponds to the CUDA initialization overhead.
        3. The final gap starting after the last GPU operation, which
           corresponds to the flushing of CUPTI buffers when stopping the collection.
"""

# This query gives the final result.
# The name of device is printed instead of the ID.
    query_final = """
WITH
    gaps AS (
        SELECT
            gapend - gapstart AS duration,
            *
        FROM
            ({GAPS})
    ),
    gpuinfo AS (
        SELECT
            name || ', ' || pciBusId AS name,
            deviceId
        FROM
            TARGET_INFO_CUDA_GPU
        GROUP BY
            deviceId
    )
    SELECT
        ROW_NUMBER() OVER(ORDER BY duration DESC) AS "Row#",
        duration AS "Duration (ns)",
        gapstart AS "Start Time (ns)",
        pid AS "Pid",
        gpuinfo.name AS "Device ID"
    FROM
        gaps
    JOIN
        gpuinfo
    ON
        gpuinfo.deviceId = gaps.deviceId
    ORDER BY
        duration DESC, gapstart
    LIMIT {ROW_LIMIT}
"""

# This query finds gaps.
# ops is the table containing kernel + memory + CUDA profiling overhead.
# 1. Subquery "starts": Give a rowNum, SRi, to each start (ordered by start time).
# 2. Subquery "ends": Give a rowNum, ERj, to each end (ordered by end time).
# 3. Reconstruct intervals [ERj, SRj+1] by putting together an end ERj from ends
#    with the next start SRj+1 from starts i.e. start_rowNum - 1 = end_rowNum.
#    For example, if an end has 2 as rowNum it will be joined with
#    a start that has 3 as rowNum.
# 4. Keep only those intervals [ERj, SRj+1] that are valid i.e. ERj < SRj+1.
# Assume that we have the following intervals:
#
# SR1                          ER2
#  |--------------a-------------|
#      SR2                ER1
#       |---------b--------|
#                                         SR3              ER3
#                                          |--------c-------|
# With step 3, we get:
# 1. ER1 joined with SR2.
# 2. ER2 joined with SR3.
#
#      SR2                 ER1
#       |---------a'--------|
#                               ER2        SR3
#                                |----b'----|
#
# However, only the second interval (b') meets the condition end < start of step 4 and
# will be considered as a gap.
# The first one (a') will be discared and the query will return:
#
#                               ER2        SR3
#                                |----b'----|
#
# ER2 will be the start and SR3 will be the end of the gap.

    query_gap = """
WITH
    ops AS (
        {OPS_ALL}
    ),
    starts AS (
        SELECT
            ROW_NUMBER() OVER(ORDER BY pid, deviceId, start) AS rowNum,
            start,
            deviceId,
            pid
        FROM
            ops
    ),
    ends AS (
        SELECT
            ROW_NUMBER() OVER(ORDER BY pid, deviceId, end) AS rowNum,
            end,
            deviceId,
            pid
        FROM
            ops
    )
    SELECT
        end AS gapstart,
        start AS gapend,
        starts.deviceId,
        starts.pid
    FROM
        starts
    JOIN
        ends
    ON
        starts.rowNum - 1 = ends.rowNum
    WHERE
        start - end > {THRESHOLD} * 1000000
        AND starts.deviceId = ends.deviceId
        AND starts.pid = ends.pid
        AND end < start
"""

# This query is used for kernel/memory operations. It selects columns
# that are necessary for finding gaps.
    query_select = """
    SELECT
        start,
        end,
        deviceId,
        ((globalPid >> 24) & 0x00FFFFFF) AS pid
    FROM
        {GPU_OPERATION}
"""

# This query is used to combine kernel/memory operations.
    query_union = """
        UNION ALL
"""

# This query adds the profiler overhead to the table that has all the
# kernel/memory operations (returned by "query_union").
# 1. Subquery "cudaoverhead": Select CUDA profiling overhead that we want
#    to take into account
# 2. Subquery "gpu": Get GPU info per deviceId/PID. [min(start), max(end)] is
#    going to be the clipping range for overheads.
# 3. Duplicate overhead rows for every deviceId that has the same PID.
#    In this way, we can have a deviceId column for overheads as well.
#    i.e., a CUDA profiling overhead on one thread affects all GPUs
#    of the same process.
# 4. The overhead rows are combined with GPU operation rows.
    query_overhead = """
    WITH
    cudaoverheadID AS (
        SELECT
            id
        FROM
            StringIds
        WHERE
            value = 'CUDA profiling data flush overhead'
            OR value = 'CUDA profiling stop overhead'
            OR value = 'CUDA profiling overhead'
    ),
    cudaoverhead AS (
        SELECT
            start,
            end,
            ((globalTid >> 24) & 0x00FFFFFF) AS pid
        FROM
            PROFILER_OVERHEAD AS po
        JOIN
            cudaoverheadID AS co
        ON co.id = po.nameId
    ),
    gpu AS (
        SELECT
            min(start) AS start,
            max(end) AS end,
            deviceId,
            pid
        FROM
            ({GPU_OPS_ALL})
        GROUP BY deviceId, pid
    )
    SELECT
        co.start,
        co.end,
        gpu.deviceId,
        co.pid
    FROM
        cudaoverhead AS co
    JOIN
        gpu
    ON
        co.pid = gpu.pid
        AND co.start > gpu.start
        AND co.end < gpu.end
    UNION ALL
    SELECT
        *
    FROM
        ({GPU_OPS_ALL})
"""

    def setup(self):
        err = super().setup()
        if err != None:
            return err

        if len(self.args) > 0 and self.args[0].isdigit():
            row_limit = self.args[0]
        else:
            row_limit = self.ROW_LIMIT

        if len(self.args) > 1 and self.args[1].isdigit():
            threshold = self.args[1]
        else:
            threshold = self.THRESHOLD

        sub_queries = []

        kernel = 'CUPTI_ACTIVITY_KIND_KERNEL'
        memcpy = 'CUPTI_ACTIVITY_KIND_MEMCPY'
        memset = 'CUPTI_ACTIVITY_KIND_MEMSET'
        overhead = 'PROFILER_OVERHEAD'

        if self.table_exists(kernel):
            sub_queries.append(self.query_select.format(GPU_OPERATION = kernel))

        if self.table_exists(memcpy):
            sub_queries.append(self.query_select.format(GPU_OPERATION = memcpy))

        if self.table_exists(memset):
            sub_queries.append(self.query_select.format(GPU_OPERATION = memset))

        if len(sub_queries) == 0:
            return 'There were no problems detected because the report does' \
                'not contain GPU kernel/memory operations data.'

        union = self.query_union.join(sub_queries)

        if self.table_exists(overhead):
            union = self.query_overhead.format(GPU_OPS_ALL = union)

        gaps = self.query_gap.format(
            OPS_ALL = union,
            THRESHOLD = threshold
        )

        self.query = self.query_final.format(
            GAPS = gaps,
            ROW_LIMIT = row_limit
        )

if __name__ == "__main__":
    CUDAGPUStarvation.Main()
